@@ -22,11 +22,6 @@
  * PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include <fontconfig/fontconfig.h>
-#include <stdio.h>
-#include <unistd.h>
-#include <stdlib.h>
-#include <string.h>
 #ifdef HAVE_CONFIG_H
 #include <config.h>
 #else
@@ -36,6 +31,12 @@
 #define HAVE_GETOPT 1
 #endif
 
+#include <fontconfig/fontconfig.h>
+#include <stdio.h>
+#include <unistd.h>
+#include <stdlib.h>
+#include <string.h>
+
 #ifndef HAVE_GETOPT
 #define HAVE_GETOPT 0
 #endif
@@ -47,8 +48,9 @@
 #undef  _GNU_SOURCE
 #define _GNU_SOURCE
 #include <getopt.h>
-const struct option longopts[] = {
+static const struct option longopts[] = {
     {"sort", 0, 0, 's'},
+    {"all", 0, 0, 'a'},
     {"version", 0, 0, 'V'},
     {"verbose", 0, 0, 'v'},
     {"help", 0, 0, '?'},
@@ -64,7 +66,7 @@ extern int optind, opterr, optopt;
 static void usage (char *program)
 {
 #if HAVE_GETOPT_LONG
-    fprintf (stderr, "usage: %s [-svV?] [--sort] [--verbose] [--version] [--help] [pattern]\n",
+    fprintf (stderr, "usage: %s [-svV?] [--sort] [--all] [--verbose] [--version] [--help] [pattern]\n",
 	     program);
 #else
     fprintf (stderr, "usage: %s [-svV?] [pattern]\n",
@@ -74,11 +76,13 @@ static void usage (char *program)
     fprintf (stderr, "\n");
 #if HAVE_GETOPT_LONG
     fprintf (stderr, "  -s, --sort           display sorted list of matches\n");
+    fprintf (stderr, "  -a, --all            display unpruned sorted list of matches\n");
     fprintf (stderr, "  -v, --verbose        display entire font pattern\n");
     fprintf (stderr, "  -V, --version        display font config version and exit\n");
     fprintf (stderr, "  -?, --help           display this help and exit\n");
 #else
     fprintf (stderr, "  -s,        (sort)    display sorted list of matches\n");
+    fprintf (stderr, "  -a         (all)     display unpruned sorted list of matches\n");
     fprintf (stderr, "  -v         (verbose) display entire font pattern\n");
     fprintf (stderr, "  -V         (version) display font config version and exit\n");
     fprintf (stderr, "  -?         (help)    display this help and exit\n");
@@ -90,7 +94,7 @@ int
 main (int argc, char **argv)
 {
     int		verbose = 0;
-    int		sort = 0;
+    int		sort = 0, all = 0;
     int		i;
     FcFontSet	*fs;
     FcPattern   *pat;
@@ -99,12 +103,15 @@ main (int argc, char **argv)
     int		c;
 
 #if HAVE_GETOPT_LONG
-    while ((c = getopt_long (argc, argv, "sVv?", longopts, NULL)) != -1)
+    while ((c = getopt_long (argc, argv, "asVv?", longopts, NULL)) != -1)
 #else
-    while ((c = getopt (argc, argv, "sVv?")) != -1)
+    while ((c = getopt (argc, argv, "asVv?")) != -1)
 #endif
     {
 	switch (c) {
+	case 'a':
+	    all = 1;
+	    break;
 	case 's':
 	    sort = 1;
 	    break;
@@ -134,21 +141,39 @@ main (int argc, char **argv)
     else
 	pat = FcPatternCreate ();
 
+    if (!pat)
+	return 1;
+
     FcConfigSubstitute (0, pat, FcMatchPattern);
     FcDefaultSubstitute (pat);
     
-    if (sort)
-	fs = FcFontSort (0, pat, FcTrue, 0, &result);
+    fs = FcFontSetCreate ();
+
+    if (sort || all)
+    {
+	FcFontSet	*font_patterns;
+	int	j;
+	font_patterns = FcFontSort (0, pat, all ? FcFalse : FcTrue, 0, &result);
+
+	for (j = 0; j < font_patterns->nfont; j++)
+	{
+	    FcPattern  *font_pattern;
+
+	    font_pattern = FcFontRenderPrepare (NULL, pat, font_patterns->fonts[j]);
+	    if (font_pattern)
+		FcFontSetAdd (fs, font_pattern);
+	}
+
+	FcFontSetSortDestroy (font_patterns);
+    }
     else
     {
 	FcPattern   *match;
-	fs = FcFontSetCreate ();
 	match = FcFontMatch (0, pat, &result);
 	if (match)
 	    FcFontSetAdd (fs, match);
     }
-    if (pat)
-	FcPatternDestroy (pat);
+    FcPatternDestroy (pat);
 
     if (fs)
     {
@@ -177,12 +202,13 @@ main (int argc, char **argv)
 		if (FcPatternGetString (fs->fonts[j], FC_FAMILY, 0, &family) != FcResultMatch)
 		    family = (FcChar8 *) "<unknown family>";
 		if (FcPatternGetString (fs->fonts[j], FC_STYLE, 0, &style) != FcResultMatch)
-		    file = (FcChar8 *) "<unknown style>";
+		    style = (FcChar8 *) "<unknown style>";
 
 		printf ("%s: \"%s\" \"%s\"\n", file, family, style);
 	    }
 	}
 	FcFontSetDestroy (fs);
     }
+    FcFini ();
     return 0;
 }
